import os
import pathlib
import re
import sys
from concurrent.futures import ThreadPoolExecutor
from configparser import ConfigParser
from typing import List, Optional, Tuple, Union

import requests
import yaml
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TransferSpeedColumn,
)

# What west's APIs accept for paths.
#
# Here, os.PathLike objects should return str from their __fspath__
# methods, not bytes. We could try to do something like the approach
# taken in https://github.com/python/mypy/issues/5264 to annotate that
# as os.PathLike[str] if TYPE_CHECKING and plain os.PathLike
# otherwise, but it doesn't seem worth it.
PathType = Union[str, os.PathLike]

# Used to keep track of supported python versions
supported_python_versions = [
    {"major": 3, "minor": 8},
    {"major": 3, "minor": 9},
]

progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}"),
    BarColumn(),
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
)


class WestDirNotFound(RuntimeError):
    """Neither the current directory nor any parent has a west workspace."""


class WestConfigNotFound(RuntimeError):
    """.west/config file does not exist."""


class WestYmlNotFound(RuntimeError):
    """west.yml file does not exist."""


def check_python_version(east):
    """Checks if current python version is supported. If not it exists with an error
    message."""

    current_py_ver = {
        "major": sys.version_info.major,
        "minor": sys.version_info.minor,
    }

    if current_py_ver not in supported_python_versions:
        vers = supported_python_versions

        east.print(
            f"You are running Python {sys.version.split(' ')[0]} which is not"
            " supported.\n"
            "Supported versions are:",
            end="",
        )

        # Nicely print a list of supported python version in markdown
        vers_str = [f"- v{ver['major']}.{ver['minor']}.x" for ver in vers]
        east.print_markdown("\n".join(vers_str))
        east.exit()


def download_file(task_id: TaskID, url: str, path: str):
    file_size = requests.head(url, allow_redirects=True).headers.get(
        "content-length", -1
    )
    response = requests.get(url, stream=True)

    progress.update(task_id, total=int(file_size))

    with open(path, "wb") as f:
        progress.start_task(task_id)
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                progress.update(task_id, advance=len(chunk))
                f.write(chunk)


def download_files(urls: List[str], dest_dir: str) -> List[str]:
    """Download concurrently multiple files from the internet to the given directory.

    Function expects a list of urls that point to the files.

    After all files were downloaded the function returns a list of paths to the
    downloaded files in the same order as they were given.

    Downloaded files are not renamed, they have the same name as in the url. Only
    exception to this rule are raw files from the GitHub, they end with '?raw=true', so
    that part is removed.

    Args:
        urls (List[strl]):      Url that points to a file to be downloaded.

    Return:
        files (List[str]):      List of paths to the downloaded files.
    """

    file_paths = []

    with progress:
        with ThreadPoolExecutor(max_workers=4) as pool:
            for url in urls:
                filename = url.split("/")[-1]
                filename = re.sub("\?raw=true$", "", filename)

                if not os.path.isdir(dest_dir):
                    os.mkdir(dest_dir)

                dest_path = os.path.join(dest_dir, filename)
                file_paths.append(dest_path)

                task_id = progress.add_task("download", filename=filename, start=False)
                pool.submit(download_file, task_id, url, dest_path)

    return file_paths


def west_topdir(start: Optional[PathType] = None) -> str:
    """
    Returns the path to the parent directory of the .west/
    directory instead, where project repositories are stored.

    Args:
        start (Optional[PathType]):     Directory from where to start searching, if not
                                        given current directory is used.

    Returns:
        Full path to parent directory of the .west/ folder.
    """
    cur_dir = pathlib.Path(start or os.getcwd())

    while True:
        if (cur_dir / ".west").is_dir():
            return os.fspath(cur_dir)

        parent_dir = cur_dir.parent
        if cur_dir == parent_dir:
            # We are at top level
            raise WestDirNotFound(
                "Could not find a west workspace in this or any parent directory"
            )
        cur_dir = parent_dir


def get_ncs_and_project_dir(west_dir_path: str) -> Tuple[str, str]:
    """
    Returns version of nrf-sdk project and absolute path to the projects directory.

    This is combined, so we avoid reading .west/config file twice.

    Args:
        west_dir_path (str):        Path to the parent of .west directory. It is up to
                                    the caller to provide a correct value.

    Returns:
        Revision string of nrf-sdk project, absolute

    Raises
        WestConfigNotFound or WestYamlNotFound
    """

    config = ConfigParser()
    config_path = os.path.join(west_dir_path, ".west", "config")

    # ".west/config file could not exist, in that case we should raise an exception"
    if not os.path.isfile(config_path):
        raise WestConfigNotFound(".west/conifg file does not exists.")

    config.read(config_path)

    # Get path to west.yaml file
    project_path = os.path.join(west_dir_path, config["manifest"]["path"])
    west_yaml = os.path.join(project_path, config["manifest"]["file"])

    if not os.path.isfile(west_yaml):
        raise WestYmlNotFound("No west.yml was found ")

    # Get ncs version
    with open(west_yaml, "r") as file:
        manifest = yaml.safe_load(file)["manifest"]

    try:
        ncs = list(
            filter(
                lambda project: project["repo-path"] == "sdk-nrf", manifest["projects"]
            )
        )
    except KeyError:
        # This can happen in the case where there is no sdk-nrf repo in the west yaml
        # file, project is probably using ordinary Zephyr.
        return None, project_path

    return (ncs[0]["revision"], project_path)


def return_dict_on_match(array_of_dicts, key, value):
    """
    Search through array of dicts and return the first one where the given key matches
    the given value.
    If none are found return None.
    """
    return next((item for item in array_of_dicts if item.get(key) == value), None)


no_toolchain_manager_msg = """
[bold cyan]Nordic's Toolchain Manager[/] is [bold red]not installed[/] on this system!

To install it run:

\t[italic bold blue]east sys-setup
"""

not_in_west_workspace_msg = """
[bold yellow]West workspace[/] was [bold red]not found![/]

This command can only be run [bold]inside[/] of a [bold yellow]West workspace[/].
"""


def no_toolchain_msg(east):
    return (
        f"Current [bold cyan]NCS[/] [bold]{east.detected_ncs_version}[/] version is "
        "supported but toolchain is [bold red]not installed![/]"
        "\n\nTo install it run:"
        "\n\n\t[italic bold blue]east update toolchain\n"
    )


def ncs_version_not_supported_msg(east, supported_versions):
    vers = "\n".join(
        [f"[bold yellow]•[/] {ver}" for ver in supported_versions.strip().split("\n")]
    )

    return (
        f"[bold]East[/] detected [bold]{east.detected_ncs_version}[/] [bold cyan]NCS[/]"
        " version which is currently [bold red]not supported[/] by the Nordic's"
        " Toolchain manager.\n\nSupported versions are: \n"
        + vers
        + "\n\nThis means that you need to manually install the [bold cyan]NCS[/]"
        " toolchain by yourself.\n"
    )


def find_all_boards(east, west_board: str) -> List[str]:
    """Find all west board names by searching the boards directory

    Search for directory that contains *_defconfig file with given west_board name,
    (this is the same process that west uses to determine the boards).

    After such directory is found, scan it and try to find what board revisions are
    present and generate a list of board names and board revisions in format expected by
    west build command.

    If no folder is found or there are no revision specific files in the folder then
    just return west_board.

        east ():            East context.
        west_board (str):   Board to search for.

    Returns:
        List of west boards to be used by west build -b <board>
    """

    def dir_find(root, west_board):
        for path, _, files in os.walk(root):
            for file in files:
                if file.endswith("_defconfig"):
                    if west_board == file[: -len("_defconfig")]:
                        return path
        return None

    board_dir = dir_find(os.path.join(east.project_dir, "boards"), west_board)

    if not board_dir:
        return [west_board]

    files = os.listdir(board_dir)

    # Find all files with west board name, no matter the version
    pattern = f"^{west_board}_[0-9]?[0-9]?_[0-9]?[0-9]?_[0-9]?[0-9]?\.conf"
    matches = [file for file in files if re.match(pattern, file)]

    # Extract hardware versions and put them into format expected by the west.
    hw_versions = [
        ".".join(m.split(".")[0].replace(west_board, "").split("_")[1:])
        for m in matches
    ]
    boards = ["@".join([west_board, hw]) for hw in sorted(hw_versions)]

    return boards if boards else [west_board]



import os

# From this file only dict consts_path should be imported, and this should happen only
# in __main__.
# There is however one exception: NRF_TOOLCHAIN_MANAGER_PATH is needed for downloading
# in sys_setup.py.

HOME_DIR = os.path.join("/home", os.environ["USER"])

# Directory that can be used for temporary intermediate files that we do not care about.
CACHE_DIR = os.path.join(HOME_DIR, ".cache", "east")

# East directory, any files that east needs to function normally are located here
EAST_DIR = os.path.join(HOME_DIR, ".local", "share", "east")

# Path to the toolchain executable, ignore the .exe extension, this works as it should
NRF_TOOLCHAIN_MANAGER_PATH = os.path.join(EAST_DIR, "nrfutil-toolchain-manager.exe")

# Directory will all Conda stuff
MINICONDA_DIR = os.path.join(HOME_DIR, "miniconda3")

# Path to the Conda executable, this can be used when the conda is not yet on PATH
CONDA_PATH = os.path.join(MINICONDA_DIR, "bin", "conda")

const_paths = {
    "home_dir": HOME_DIR,
    "cache_dir": CACHE_DIR,
    "east_dir": EAST_DIR,
    "nrf_toolchain_manager_path": NRF_TOOLCHAIN_MANAGER_PATH,
    "miniconda_dir": MINICONDA_DIR,
    "conda_path": CONDA_PATH,
}



import click
import rich_click

from .east_context import EastContext, east_group_settings
from .system_commands import sys_setup, util
from .workspace_commands import build, bypass, clean, flash, release, update

rich_click.rich_click.MAX_WIDTH = 80
rich_click.rich_click.USE_RICH_MARKUP = True
rich_click.rich_click.STYLE_HELPTEXT = "italic dim"

rich_click.rich_click.COMMAND_GROUPS = {
    "east": [
        {
            "name": "Workspace commands",
            "commands": ["build", "flash", "clean", "update", "bypass", "release"],
        },
        {
            "name": "System commands",
            "commands": ["sys-setup", "util"],
        },
    ],
    "east update": [{"name": "Subcommands", "commands": ["west", "env", "toolchain"]}],
    "east util": [{"name": "Subcommands", "commands": ["connect", "rtt"]}],
}


@click.group(
    **east_group_settings,
    chain=False,
    subcommand_metavar="<command> [command options]",
)
@click.version_option(message="%(version)s", package_name="east-tool")
@click.option(
    "--echo", is_flag=True, help="Echo each shell command before executing it."
)
@click.pass_context
def cli(ctx, echo):
    """
    [bold]East[/] is a command line meta-tool, useful for creating, managing and
    deploying [bold cyan]nRF Connect SDK[/] projects.

    \b
    \n\nWant to learn what each command does?

    Run [bold]east \[command] --help[/] to show documentation for that command.

    \b
    \n\nNote that commands are split into two groups:

    - [bold]Workspace:[/] Can only be run from inside of [bold yellow]West workspace[/].

    - [bold]System:[/] Can be run from anywhere.
    """

    # EastContext object is passed to other subcommands due to the @click.pass_context
    # decorator. Additionally, the subcommands need to be decorated with @click.pass_obj
    # so they directly access the EastContext object.
    ctx.obj = EastContext(echo)
    pass


cli.add_command(build)
cli.add_command(bypass)
cli.add_command(flash)
cli.add_command(clean)
cli.add_command(update)
cli.add_command(sys_setup)
cli.add_command(util)
cli.add_command(release)


def main():
    cli()


if __name__ == "__main__":
    main()



import os
import re
import sys

import click

from ..constants import NRF_TOOLCHAIN_MANAGER_PATH
from ..east_context import EastContext, east_command_settings
from ..helper_functions import check_python_version, download_files


def _get_conda_download_link():
    """Construct download link for conda installer script based on system python
    version"""

    # Link to the conda installer script. Two big Xs will be later replaced with python
    # version, like 39 or 38.
    link = "https://repo.anaconda.com/miniconda/Miniconda3-pyXX_4.12.0-Linux-x86_64.sh"
    py_ver = str(sys.version_info.major) + str(sys.version_info.minor)
    return re.sub("XX", py_ver, link)


def _install_conda(east: EastContext, installer_path: str):
    """Installs Conda package manager"""

    # Conda is not on path, but the miniconda dir could still exist.
    # If we do not delete it now the installer will complain.
    east.run(f"rm -fr {east.consts['miniconda_dir']}")

    east.print("[bold blue]Started Conda installer...")
    # -b flag stands for '[b]e silent', hahaha....
    east.run(f"bash {installer_path} -b ")

    # Conda is now installed, but not on the path yet (this will happen after
    # sourcing .bashrc, .zshrc, .fishrc, etc..), so we talk to it with fullpath
    # Do not activate base env.
    east.run(f"{east.consts['conda_path']} config --set auto_activate_base false")
    east.run(f"{east.consts['conda_path']} init")


def _install_toolchain_manager(east: EastContext, exe_path: str):
    """Installs toolchain manager to a proper location"""

    # Move toolchain manager to its proper place
    east.run(f"mv -f {exe_path} {east.consts['east_dir']}")

    # That is octal, make it executable
    os.chmod(east.consts["nrf_toolchain_manager_path"], 0o777)

    # Configure the toolchain path
    east.run(
        f"{east.consts['nrf_toolchain_manager_path']} config --install-dir "
        f"{east.consts['east_dir']}"
    )


def _get_toolchain_download_link():
    """Just a convenience function for getting the link for the Nordic's toolchain
    manager executable.

    This link will be updated regularly to follow the progress on the Nordic's repo.

    Current version of the executable is v0.8.0.
    """

    link = (
        "https://github.com/NordicSemiconductor/pc-nrfconnect-toolchain-manager/"
        "blob/2f24ef572b8a7182cb6838fc2f080ad1b4fee448/resources/"
        "nrfutil-toolchain-manager/linux/nrfutil-toolchain-manager.exe?raw=true"
    )
    return link


conda_installed_msg = """
[bold green]Conda install done![/]

Close and reopen your terminal window or source your shell configuration file
manually with one of the below commands for changes to take effect:
• source ~/.bashrc
• source ~/.zshrc
• source ~/.config/fish/config.fish
"""

toolchain_installed_msg = """
[bold green]Nordic's Toolchain Manager install done![/]

East will now smartly use Nordic's Toolechain Manager whenever it can.

[bold]Note:[/] You still need to run [italic bold blue]east update toolchain[/] inside
of a [yellow bold]West workspace[/] to get the actual toolchain.
"""

packages = [
    {
        "exe": "conda",
        "url": _get_conda_download_link(),
        "install_method": _install_conda,
        "installed_msg": conda_installed_msg,
    },
    {
        "exe": NRF_TOOLCHAIN_MANAGER_PATH,
        "url": _get_toolchain_download_link(),
        "install_method": _install_toolchain_manager,
        "installed_msg": toolchain_installed_msg,
    },
]


@click.command(**east_command_settings)
@click.pass_obj
def sys_setup(east):
    """Perform system-wide setup for development.

    \b
    \n\nCheck if all required packages are available on the host system.

    If not, download and install them.

    \b
    \n\nPackages:\n
    - Conda Package Manager

    - Nordic's nRF Toolchain Manager executable
    """

    check_python_version(east)

    # Construct a list of files that have to be downloaded.
    east.print(
        "[blue]Checking for required system packages and programs...", highlight=False
    )
    urls = []
    for package in packages:
        package["installed"] = east.check_exe(package["exe"])
        if package["installed"]:
            east.print(f"{package['exe']} [green]found", highlight=False)
        else:
            east.print(f"{package['exe']} [red]not found", highlight=False)
            urls.append(package["url"])

    if all([package["installed"] is True for package in packages]):
        east.print("\n[green]All required system packages and programs are installed.")
        east.exit(0)

    # Download all required files, which are actually programs or installer scripts
    paths = download_files(urls, east.consts["cache_dir"])

    # Run an installation method for packages that are not installed.
    for package in packages:
        if not package["installed"]:
            index = urls.index(package["url"])
            package["install_method"](east, paths[index])

    # WARN: We assume that download_files will never fail. Depending on the urgency we
    # should implement better handling.
    # Print installed message for packages that were installed
    for package in packages:
        if not package["installed"]:
            east.console.rule("", style="")
            east.print(package["installed_msg"])
    east.console.rule("", style="")


@click.command(**east_command_settings)
@click.pass_obj
def init(east):
    """Creates a West workspace."""



import os
import shutil as sh

import click

from ..east_context import east_command_settings
from .build_type_flag import construct_extra_cmake_arguments


@click.command(**east_command_settings)
@click.pass_obj
def clean(east):
    """Clean the build folder in current directory.



    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    east.run("rm -fr build")


@click.command(**east_command_settings)
@click.option("-b", "--board", type=str, help="West board to build for.")
@click.option(
    "-u",
    "--build-type",
    type=str,
    help=(
        "Which build type (a group of [bold]Kconfig[/] fragment files) to use. Requires"
        " [bold yellow]east.yml[/] with possible build types."
    ),
)
@click.option(
    "-d",
    "--build-dir",
    type=str,
    help=(
        "Build directory to create or use. If the --build-dir directory is not set, the"
        " default is [bold]build[/] unless the build.dir-fmt configuration variable is"
        " set. The current directory is checked after that. If either is a Zephyr build"
        " directory, it is used. "
    ),
)
@click.option("-t", "--target", type=str, help="Run this build system target.")
# @click.argument("cmake-args", nargs=-1, type=str, metavar="-- [cmake-args]")
@click.option(
    "-s",
    "--source-dir",
    type=str,
    help=(
        "Relative path to a directory that should be used as application source"
        " directory."
    ),
)
@click.pass_obj
def build(east, board, build_type, build_dir, target, source_dir):
    """
    Build firmware in the current directory.

    \b
    \n\nInternally runs [magenta bold]west build[/] command in current directory if --source-dir is not set.

    \n\nTo pass additional arguments to the [bold]CMake[/] invocation performed by the [magenta
    bold]west build[/], pass them after a [bold white]"--"[/] at the end of the command line.

    \n\n[bold]Important:[/] Passing additional [bold]CMake[/] arguments like this forces [magenta
    bold]west build[/] to re-run [bold]CMake[/], even if a build system has already been generated.

    For additional info see chapter [bold]Building, Flashing and Debugging[/], section
    [bold]One-Time CMake Arguments[/].

    \n\n[bold]Note:[/] This command will, after build step, copy
    [bold]compile_commands.json[/], if found, from the build directory to the project and top west directory. This makes job of locating this file easier for [bold yellow]clangd[/].

    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """

    east.pre_workspace_command_check()

    build_cmd = create_build_command(
        east, board, build_type, build_dir, target, source_dir
    )

    east.run_west(build_cmd)

    compile_file = os.path.join("build", "compile_commands.json")
    if os.path.isfile(compile_file):
        sh.copyfile(
            compile_file, os.path.join(east.project_dir, "compile_commands.json")
        )
        sh.copyfile(
            compile_file, os.path.join(east.west_dir_path, "compile_commands.json")
        )


def create_build_command(
    east,
    board=None,
    build_type=None,
    build_dir=None,
    target=None,
    source_dir=None,
    silence_diagnostic=False,
):
    """Creates build command. This is needed so it can also be reused by release command"""
    build_cmd = "build"

    if board:
        build_cmd += f" -b {board}"
    if build_dir:
        build_cmd += f" -d {build_dir}"
    if target:
        build_cmd += f" -t {target}"
    if source_dir:
        build_cmd += f" {source_dir}"

    # WARN: cmake args are making some problems in this form.
    # if cmake_args:
    #     build_cmd += f" -- \"{' '.join(cmake_args)}\""

    build_type_args, diagnostic = construct_extra_cmake_arguments(
        east,
        build_type,
        board,
        build_dir,
        source_dir,
    )

    if diagnostic and not silence_diagnostic:
        east.print(diagnostic)

    if build_type_args:
        build_cmd += f" -- {build_type_args}"

    return build_cmd


@click.command(**east_command_settings)
@click.option("-d", "--build-dir", type=str, help="Build directory to create or use.")
@click.option(
    "-r", "--runner", type=str, help="Override default runner from --build-dir."
)
@click.option(
    "-v",
    "--verify",
    is_flag=True,
    help=(
        "Verify, after flash, that contents of the boards code memory regions are the"
        " same as in flashed image."
    ),
)
@click.option(
    "-i",
    "--jlink-id",
    type=str,
    help=(
        "Identification number of a JLink programmer that should be used for flashing."
    ),
)
@click.argument("extra-args", nargs=-1, type=str, metavar="-- [extra args]")
@click.pass_obj
def flash(east, build_dir, runner, verify, jlink_id, extra_args):
    """
    Flash binary to the board's flash.

    \b
    \n\nInternally runs [magenta bold]west flash[/] command. If the build directory is not given, the default is build/ unless the build.dir-fmt configuration variable is set. The current directory is checked after that. If either is a Zephyr build directory, it is used. If there are more than one JLinks connected to the host machine use --jlink-id flag to specify which one to use to avoid selection prompt.

    \n\nTo pass additional arguments to the [bold cyan]runner[/] used by the [magenta
    bold]west flash[/], pass them after a [bold white]"--"[/] at the end of the command line.



    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    flash_cmd = "flash "

    if build_dir:
        flash_cmd += f"-d {build_dir} "
    if runner:
        flash_cmd += f"-r {runner} "

    # Some flags need to be passed as extra parameters to the west tool
    if verify or jlink_id or extra_args:
        flash_cmd += "-- "

    if verify:
        flash_cmd += "--verify "
    if jlink_id:
        flash_cmd += f"-i {jlink_id} "
    if extra_args:
        flash_cmd += f"\"{' '.join(extra_args)}\" "

    east.run_west(flash_cmd)


@click.command(**east_command_settings)
@click.argument("args", nargs=-1, type=str, metavar="-- [args]")
@click.pass_obj
def bypass(east, args):
    """
    Bypass any set of commands directly to the [magenta bold]west tool[/].

    \b
    \n\nPassing any set of commands after double dash [bold]--[/] will pass them directly to
    the [bold magenta]west[/] tool.

    \n\nExample:

    \n\nCommand [bold]east bypass -- build -b nrf52840dk_nrf52840[/]
    \n\nbecomes [bold]west build -b nrf52840dk_nrf52840[/]



    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    if not args:
        east.exit()

    # Click argument automatically strips double quotes from anything that is given
    # after "--". Double quotes are needed if specifying define values (-D) to the cmake
    # args, below list comprehension adds them back.
    def add_back_double_quotes(arg):
        splited = arg.split("=")
        return f'{splited[0]}="{splited[1]}"'

    args = [add_back_double_quotes(arg) if "=" in arg else arg for arg in args]

    cmd = f"{' '.join(args)} "
    east.run_west(cmd)



import os

from click.testing import CliRunner

import east
from east.__main__ import cli
from east.east_context import EastContext
from east.helper_functions import find_all_boards

from . import helpers
from .helpers import helper_test_against_west_run1


def helper_test_against_west_run(
    monkeypatch, mocker, path, east_cmd, expected_west_cmds=None, should_succed=True
):
    """
    Helper function for making tests easier to read.

    Args:
        monkeypatch ():         fixture
        mocker ():              fixture
        path ():                To which path should we change
        east_cmd ():            which east command should be called
        expected_west_cmds ():  List of expected west_cmds. If
                                none then no run_west call should happend.
        should_succed ():       If true then the command should succeded.

    Returns:
        Result object, which can be further checked.
    """
    runner = CliRunner()

    # Mock output of git commmand, so tests do not have to depend on it
    mocker.patch(
        "east.workspace_commands.release_commands.get_git_version",
        return_value={"tag": "v1.0.0.", "hash": ""},
    )

    monkeypatch.chdir(path)
    mocker.patch(
        "east.east_context.EastContext.run_west",
        return_value={"output": "", "returncode": 0},
    )

    # Setting catch_exceptions to False enables us to see programming errors in East
    # code
    result = runner.invoke(cli, east_cmd.strip().split(" "), catch_exceptions=False)

    run_west = east.east_context.EastContext.run_west

    if expected_west_cmds:
        # This conversion is needed due to assert_has_calls interface
        calls = [
            mocker.call(cmd, silent=True, return_output=True, exit_on_error=False)
            for cmd in expected_west_cmds
        ]
        run_west.assert_has_calls(calls)
    else:
        run_west.assert_not_called()

    expected_return_code = 0 if should_succed else 1

    assert result.exit_code == expected_return_code
    return result


east_yaml_single_app = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
      - type: uart
        conf-files:
          - debug.conf
          - uart.conf
"""


def test_finding_hardware_versions(west_workplace):
    _ = west_workplace
    east = EastContext()

    # Some board that has as several different versions in board folder
    boards = find_all_boards(east, "custom_nrf52840dk")
    expected_boards = [
        "custom_nrf52840dk@1.0.0",
        "custom_nrf52840dk@1.1.0",
        "custom_nrf52840dk@2.20.1",
    ]
    assert expected_boards == boards

    # Some board that has a board folder but not revisions
    boards = find_all_boards(east, "nrf52840dk_nrf52840")
    expected_boards = ["nrf52840dk_nrf52840"]
    assert expected_boards == boards

    # Non-existing in sense that it does not have its own boards folder.
    boards = find_all_boards(east, "non_existing_board")
    expected_boards = ["non_existing_board"]
    assert expected_boards == boards


def test_no_east_yml_with_release(west_workplace_parametrized, monkeypatch, mocker):
    """
    Running east release with not east.yml should abort.
    """

    os.remove("east.yml")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["project"],
        "release",
        should_succed=False,
    )


east_yaml_no_samples_key = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
      - type: uart
        conf-files:
          - debug.conf
          - uart.conf
"""

expected_app_release_west_commands = [
    (
        "build -b custom_nrf52840dk@1.0.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf"'
    ),
    (
        "build -b custom_nrf52840dk@1.0.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
    ),
    "build -b custom_nrf52840dk@1.0.0 app -- -DCONF_FILE=conf/common.conf",
    (
        "build -b custom_nrf52840dk@1.1.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf"'
    ),
    (
        "build -b custom_nrf52840dk@1.1.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
    ),
    "build -b custom_nrf52840dk@1.1.0 app -- -DCONF_FILE=conf/common.conf",
    (
        "build -b custom_nrf52840dk@2.20.1 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf"'
    ),
    (
        "build -b custom_nrf52840dk@2.20.1 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
    ),
    "build -b custom_nrf52840dk@2.20.1 app -- -DCONF_FILE=conf/common.conf",
    (
        "build -b nrf52840dk_nrf52840 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;conf/debug.conf"'
    ),
    (
        "build -b nrf52840dk_nrf52840 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;conf/debug.conf;conf/uart.conf"'
    ),
    (
        "build -b nrf52840dk_nrf52840 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
    ),
]


def test_basic_app_release_behaviour(west_workplace, monkeypatch, mocker):
    """
    Running east release with samples key, should skip build process for samples and build
    apps.
    """

    project_path = west_workplace

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yaml_no_samples_key,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "release",
        expected_west_cmds=expected_app_release_west_commands,
    )


east_yaml_no_apps_key = """
samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
"""

expected_samples_release_west_commands = [
    "build -b custom_nrf52840dk@1.0.0 samples/settings",
    "build -b custom_nrf52840dk@1.1.0 samples/settings",
    "build -b custom_nrf52840dk@2.20.1 samples/settings",
    "build -b nrf52840dk_nrf52840 samples/settings",
    "build -b custom_nrf52840dk@1.0.0 samples/dfu",
    "build -b custom_nrf52840dk@1.1.0 samples/dfu",
    "build -b custom_nrf52840dk@2.20.1 samples/dfu",
    "build -b nrf52840dk_nrf52840 samples/dfu",
]


def test_basic_samples_release_behaviour(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    Running east release with no apps key, should skip build process for apps and build
    samples.
    """

    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_apps_key,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["project"],
        "release",
        expected_west_cmds=expected_samples_release_west_commands,
    )


def test_basic_release_behaviour(west_workplace, monkeypatch, mocker):
    """
    Running east release with both apps and samples keys should run normally.
    """
    project_path = west_workplace

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yaml_no_apps_key + east_yaml_no_samples_key,
    )
    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "release",
        expected_west_cmds=expected_app_release_west_commands
        + expected_samples_release_west_commands,
    )


east_yaml_non_existing_sample = """
samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

  - name: non_existing_sample_name
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
"""


def test_east_yml_with_non_existant_samples(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    Running east release when east.yml contains non-existing samples should abort before
    issuing any of the west build commands.
    """

    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_non_existing_sample,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["project"],
        "release",
        should_succed=False,
    )


east_yaml_non_existing_app = """
apps:
  - name: test_one
    west-boards:
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

  - name: test_two
    west-boards:
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

  - name: test_three_which_does_not_exist
    west-boards:
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
"""


def test_east_yml_with_non_existant_apps(west_workplace_multi_app, monkeypatch, mocker):
    """
    Running east release when east.yml contains non-existing apps should abort before
    issuing any of the west build commands. This only make sense in multi app
    workspaces.
    """
    project = west_workplace_multi_app

    helpers.create_and_write(
        project,
        "east.yml",
        east_yaml_non_existing_app,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project,
        "release",
        should_succed=False,
    )



import os

import pytest
from click.testing import CliRunner

import east
from east.__main__ import cli

from . import helpers


def helper_test_against_west_run(
    monkeypatch, mocker, path, east_cmd, expected_west_cmd=None, should_succed=True
):
    """
    Helper function for making tests easier to read.

    Args:
        monkeypatch ():         fixture
        mocker ():              fixture
        path ():                To which path should we change
        east_cmd ():            which east command should be called
        expected_west_cmd ():   what is expected west cmd that should be produced. If
                                none then no run_west call should happend.
        should_succed ():       If true then the command should succeded.

    Returns:
        Result object, which can be further checked.
    """
    runner = CliRunner()

    monkeypatch.chdir(path)
    mocker.patch("east.east_context.EastContext.run_west")

    # Setting catch_exceptions to False enables us to see programming errors in East
    # code
    result = runner.invoke(cli, east_cmd.strip().split(" "), catch_exceptions=False)

    run_west = east.east_context.EastContext.run_west

    if expected_west_cmd:
        run_west.assert_called_once_with(expected_west_cmd)
    else:
        run_west.assert_not_called()

    expected_return_code = 0 if should_succed else 1

    assert result.exit_code == expected_return_code
    return result


def test_build_type_with_no_east_yml(west_workplace_parametrized, monkeypatch):
    """
    --build-type option can not be used if east.yml is not found in the project root.
    East needs to exit with error and message.
    """

    os.remove("east.yml")

    monkeypatch.chdir(west_workplace_parametrized["app"])
    runner = CliRunner()

    east_cmd = "build --build-type debug".split(" ")

    # Call from the project root
    result = runner.invoke(cli, east_cmd)
    assert result.exit_code == 1


# if in project but not in app or samples then --build-type should not be given,
# we are assuming that user might be trying to run build on tests or zephyr's samples or
# something else
def test_build_type_outside_project_dir(west_workplace, monkeypatch):
    """
    No --build-type should be given if east is called:
    - inside of project dir, but not in app
    - outside of project dir
    East needs to exit with error and message.
    """

    project_path = west_workplace
    runner = CliRunner()

    east_cmd = "build --build-type debug".split(" ")

    # Call from the project root
    result = runner.invoke(cli, east_cmd)
    assert result.exit_code == 1

    # Go one level up from project dir
    monkeypatch.chdir(os.path.dirname(project_path))
    result = runner.invoke(cli, east_cmd)
    assert result.exit_code == 1

    # Go into zephyr folder
    monkeypatch.chdir(os.path.join(os.path.dirname(project_path), "zephyr"))
    result = runner.invoke(cli, east_cmd)
    assert result.exit_code == 1


# in single app folder behaviour
@pytest.mark.parametrize(
    "build_type_flag, cmake_arg",
    [
        ("", "-DCONF_FILE=conf/common.conf"),
        (
            "--build-type debug",
            '-DCONF_FILE=conf/common.conf -DOVERLAY_CONFIG="conf/debug.conf"',
        ),
        (
            "--build-type uart",
            (
                "-DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
            ),
        ),
    ],
)
def test_build_type_single_app_behaviour(
    west_workplace, monkeypatch, mocker, build_type_flag, cmake_arg
):
    """
    build command needs to parse the --build-type flag into appopriate command line
    arguments.

    This test case is for a single app behaviour.
    """
    project_path = west_workplace
    helper_test_against_west_run(
        monkeypatch,
        mocker,
        os.path.join(project_path, "app"),
        f"build {build_type_flag}",
        expected_west_cmd=f"build -- {cmake_arg}",
    )


@pytest.mark.parametrize(
    "multiapp, build_type_flag, cmake_arg",
    [
        ("test_one", "", "-DCONF_FILE=conf/common.conf"),
        (
            "test_one",
            "--build-type debug",
            '-DCONF_FILE=conf/common.conf -DOVERLAY_CONFIG="conf/debug.conf"',
        ),
        (
            "test_one",
            "--build-type uart",
            (
                "-DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
            ),
        ),
        ("test_two", "", "-DCONF_FILE=conf/common.conf"),
        (
            "test_two",
            "--build-type rtt",
            '-DCONF_FILE=conf/common.conf -DOVERLAY_CONFIG="conf/rtt.conf"',
        ),
        (
            "test_two",
            "--build-type debug-rtt",
            (
                "-DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf;conf/rtt.conf"'
            ),
        ),
    ],
)
def test_build_type_multi_app_behaviour(
    west_workplace_multi_app, monkeypatch, mocker, multiapp, build_type_flag, cmake_arg
):
    """
    build command needs to parse the --build-type flag into appopriate command line
    arguments.

    This test case is for a multi app behaviour.
    """
    project_path = west_workplace_multi_app

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        os.path.join(project_path, "app", multiapp),
        f"build {build_type_flag}",
        expected_west_cmd=f"build -- {cmake_arg}",
    )


def create_image_preload_file(
    app_path, path_prefix="", overlay_configs=None, build_dir="build"
):
    """create image_preload.cmake file inside build folder."""

    image_preload_file = (
        "# Generated file that can be used to preload variant"
        f' images\nset(CACHED_CONF_FILE "{path_prefix}conf/common.conf" CACHE INTERNAL'
        ' "NCS child image controlled")\nset(DTC_OVERLAY_FILE "" CACHE INTERNAL "NCS'
        ' child image controlled")\nset(WEST_PYTHON "/usr/bin/python3" CACHE INTERNAL'
        ' "NCS child image controlled")\n'
    )

    if overlay_configs:
        image_preload_file += (
            f'set(OVERLAY_CONFIG "{path_prefix}{overlay_configs}" CACHE INTERNAL "NCS'
            ' child image controlled")\n'
        )

    helpers.create_and_write(
        app_path, f"{build_dir}/image_preload.cmake", image_preload_file
    )


@pytest.mark.parametrize(
    "build_type_flag, overlay_configs",
    [
        ("", None),
        (
            "--build-type debug",
            "conf/debug.conf",
        ),
        (
            "--build-type uart",
            "conf/debug.conf;conf/uart.conf",
        ),
    ],
)
def test_build_type_build_folder_behaviour_same_flags(
    west_workplace_parametrized, monkeypatch, mocker, build_type_flag, overlay_configs
):
    """
    If the build folder with same conf files with that --build-type expects exits then
    no cmake args are added to the build command to avoid cmake rebuilds.
    """
    app_path = west_workplace_parametrized["app"]
    create_image_preload_file(app_path, overlay_configs=overlay_configs)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        app_path,
        f"build {build_type_flag}",
        expected_west_cmd="build",
    )


@pytest.mark.parametrize(
    "build_type_flag, overlay_configs",
    [
        (
            "--build-type debug",
            "conf/debug.conf",
        ),
        (
            "--build-type uart",
            "conf/debug.conf;conf/uart.conf",
        ),
    ],
)
def test_build_type_build_folder_behaviour_different_flags(
    west_workplace_parametrized, monkeypatch, mocker, build_type_flag, overlay_configs
):
    """
    If the build folder exsits but it has build flags that are not expected by the east
    then rebuild is triggered.
    """
    app_path = west_workplace_parametrized["app"]
    create_image_preload_file(app_path)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        app_path,
        f"build {build_type_flag}",
        expected_west_cmd=(
            "build -- -DCONF_FILE=conf/common.conf"
            f' -DOVERLAY_CONFIG="{overlay_configs}"'
        ),
    )


def test_build_type_non_existant_type(west_workplace_parametrized, monkeypatch, mocker):
    """
    If given --build-type does not exists then east needs to exit and throw message.
    """
    app_path = west_workplace_parametrized["app"]

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        app_path,
        "build --build-type asasdada",
        should_succed=False,
    )


def test_build_type_samples_with_build_type_option(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    Running east build with --build-type inside samples should fail.
    """
    project_path = west_workplace_parametrized["project"]
    sample_path = os.path.join(project_path, "samples", "settings")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build --build-type debug",
        should_succed=False,
    )


def test_build_type_samples_inherit(west_workplace_parametrized, monkeypatch, mocker):
    """
    Test inherit key normal use.
    """

    project_path = west_workplace_parametrized["project"]

    def west_cmd_fmt(prefix):
        return (
            f"build -- -DCONF_FILE={prefix}conf/common.conf"
            f' -DOVERLAY_CONFIG="{prefix}conf/debug.conf"'
        )

    sample_path = os.path.join(project_path, "samples", "settings")
    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd=west_cmd_fmt(west_workplace_parametrized["prefix"]),
        should_succed=True,
    )


def test_build_type_samples_inherit_build_folder_same_flag(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    In case where sample (with an inherit keword) has a existing build folder from
    before, no extra cmake args should be emmited.
    """
    project_path = west_workplace_parametrized["project"]

    sample_path = os.path.join(project_path, "samples", "settings")

    overlay_configs = "conf/debug.conf"
    create_image_preload_file(
        sample_path,
        path_prefix=west_workplace_parametrized["prefix"],
        overlay_configs=overlay_configs,
    )
    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd="build",
        should_succed=True,
    )


def test_build_type_samples_no_inherit(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    In case where there is no inherit we default to basic west behavior: no cmake args.
    """

    project_path = west_workplace_parametrized["project"]
    sample_path = os.path.join(project_path, "samples", "dfu")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd="build",
        should_succed=True,
    )


def test_build_type_samples_does_not_exist(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    In case where sample does not exist in east.yml we default to basic west behaviour:
    no cmake args.

    """
    project_path = west_workplace_parametrized["project"]

    sample_path = os.path.join(project_path, "samples", "super_duper_sample")
    os.mkdir(sample_path)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd="build",
        should_succed=True,
    )


east_yaml_non_existing_inheriting = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_two
        build-type: debug
"""


def test_non_existing_inherited_app(west_workplace_parametrized, monkeypatch, mocker):
    """
    In case where sample is inheriting from a non-existing app we exit.

    """

    project_path = west_workplace_parametrized["project"]

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yaml_non_existing_inheriting,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "build",
        should_succed=False,
    )


east_yaml_inheriting_from_release_single_app = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: release
"""

east_yaml_inheriting_from_release_multi_app = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

  - name: test_two
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: release
"""


def test_inheriting_from_release_build_type(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    It should be possible for a sample to inherit from a "release" build type. In that
    case only common.conf should be added to the build.
    """

    project_path = west_workplace_parametrized["project"]

    def west_cmd_fmt(prefix):
        return f"build -- -DCONF_FILE={prefix}conf/common.conf"

    if west_workplace_parametrized["project_type"] == "single":
        east_yml = east_yaml_inheriting_from_release_single_app

    if west_workplace_parametrized["project_type"] == "multi":
        east_yml = east_yaml_inheriting_from_release_multi_app

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yml,
    )

    sample_path = os.path.join(project_path, "samples", "settings")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd=west_cmd_fmt(west_workplace_parametrized["prefix"]),
        should_succed=True,
    )


east_yaml_duplicated_app_names = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
    # Don't inherit, use prj.conf in the sample's folder.
"""

east_yaml_duplicated_samples_names = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug

  - name: settings
    west-boards:
      - custom_nrf52840dk

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
    # Don't inherit, use prj.conf in the sample's folder.
"""

east_yaml_duplicated_build_types = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
      - type: debug
        conf-files:
          - debug.conf
      - type: debug
        conf-files:
          - debug.conf
      - type: rtt
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
    # Don't inherit, use prj.conf in the sample's folder.
"""


@pytest.mark.parametrize(
    "east_yml",
    [
        "east_yaml_duplicated_app_names",
        "east_yaml_duplicated_build_types",
        "east_yaml_duplicated_samples_names",
    ],
)
def test_duplicated_names_in_east_yml(
    west_workplace_multi_app, monkeypatch, mocker, east_yml
):
    """
    In case where east.yml has duplicated app names exit and throw error.
    """

    helpers.create_and_write(
        west_workplace_multi_app,
        "east.yml",
        east_yml,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_multi_app,
        "build",
        should_succed=False,
    )


@pytest.mark.parametrize(
    "east_cmd, expected_west_cmd",
    [
        (
            "build -b nrf52840dk_nrf52840",
            (
                "build -b nrf52840dk_nrf52840 -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840 --build-type uart",
            (
                "build -b nrf52840dk_nrf52840 -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;'
                'conf/debug.conf;conf/uart.conf"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840@1.0.0",
            (
                "build -b nrf52840dk_nrf52840@1.0.0 -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840@1.0.0 --build-type uart",
            (
                "build -b nrf52840dk_nrf52840@1.0.0 -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;'
                'conf/debug.conf;conf/uart.conf"'
            ),
        ),
        (
            "build -b nonexisting_board",
            "build -b nonexisting_board -- -DCONF_FILE=conf/common.conf",
        ),
        (
            "build -b nonexisting_board --build-type debug",
            (
                "build -b nonexisting_board -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf"'
            ),
        ),
    ],
)
def test_searching_for_west_board_specific_confs(
    west_workplace_parametrized,
    monkeypatch,
    mocker,
    east_cmd,
    expected_west_cmd,
):
    """
    In case where we are building for a specific board, build command needs to check in
    the conf folder if a file with name west_board.conf exists, if yes then it needs to
    add it.
    In sample folders with inherit key it needs to do the same thing.
    """
    project_path = west_workplace_parametrized["app"]

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        east_cmd,
        expected_west_cmd,
        should_succed=True,
    )


def test_different_build_dir_path_empty_dir(
    west_workplace_parametrized, monkeypatch, mocker
):
    """With different build dir path, but no build folder the expected west command
    should just include the -d option and common.conf"""

    project_path = west_workplace_parametrized["app"]

    build_dir = "../different_build_dir"

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        f"build -d {build_dir}",
        f"build -d {build_dir} -- -DCONF_FILE=conf/common.conf",
        should_succed=True,
    )


def test_different_build_dir_path_full_dir_same_build_type(
    west_workplace_parametrized, monkeypatch, mocker
):
    """With different build dir path and build folder that has same previous build type
    files as current ones the west command should just include -d option and nothing
    else.
    """

    project_path = west_workplace_parametrized["app"]

    build_dir = "../different_build_dir"

    # Config for uart build type
    overlay_configs = "conf/debug.conf;conf/uart.conf"

    create_image_preload_file(
        os.path.join(project_path, build_dir),
        path_prefix=west_workplace_parametrized["prefix"],
        overlay_configs=overlay_configs,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        f"build -d {build_dir} --build-type uart",
        f"build -d {build_dir} -- -DCONF_FILE=conf/common.conf"
        f' -DOVERLAY_CONFIG="{overlay_configs}"',
        should_succed=True,
    )


def test_different_build_dir_path_full_dir_different_build_type(
    west_workplace_parametrized, monkeypatch, mocker
):
    """With different build dir path and build folder that has different previous build
    type files as current ones the west command should include -d option and current
    conf files.
    """

    project_path = west_workplace_parametrized["app"]

    build_dir = "../different_build_dir"

    # Config for uart build type
    old_overlay_configs = "conf/debug.conf;conf/uart.conf"
    # Config for debug build type
    new_overlay_configs = "conf/debug.conf"

    create_image_preload_file(
        os.path.join(project_path, build_dir),
        path_prefix=west_workplace_parametrized["prefix"],
        overlay_configs=old_overlay_configs,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        f"build -d {build_dir} --build-type debug",
        f"build -d {build_dir} -- -DCONF_FILE=conf/common.conf"
        f' -DOVERLAY_CONFIG="{new_overlay_configs}"',
        should_succed=True,
    )


@pytest.mark.parametrize(
    "east_cmd, expected_west_cmd",
    [
        (
            "build -s app/test_one",
            "build app/test_one -- -DCONF_FILE=conf/common.conf",
        ),
        (
            "build -s app/test_two",
            "build app/test_two -- -DCONF_FILE=conf/common.conf",
        ),
        (
            "build -s app/test_one --build-type debug",
            (
                "build app/test_one -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf"'
            ),
        ),
        (
            "build -s app/test_two --build-type debug",
            (
                "build app/test_two -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf"'
            ),
        ),
        (
            "build -s app/test_one -b nrf52840dk_nrf52840",
            (
                "build -b nrf52840dk_nrf52840 app/test_one --"
                " -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
            ),
        ),
        (
            "build -s app/test_one -b nrf52840dk_nrf52840 --build-type uart",
            (
                "build -b nrf52840dk_nrf52840 app/test_one --"
                " -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;'
                'conf/debug.conf;conf/uart.conf"'
            ),
        ),
    ],
)
def test_using_different_source_dirs(
    west_workplace_multi_app,
    monkeypatch,
    mocker,
    east_cmd,
    expected_west_cmd,
):
    """
    Using a different source dir must not affect the extra args after `--`.
    """
    project_path = west_workplace_multi_app

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        east_cmd,
        expected_west_cmd,
        should_succed=True,
    )


def test_sample_with_inherit_and_with_source_dir(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    Test if sample with inherit key and source_dir works
    """

    project_path = west_workplace_parametrized["project"]

    def west_cmd_fmt(prefix):
        return (
            f"build samples/settings -- -DCONF_FILE={prefix}conf/common.conf"
            f' -DOVERLAY_CONFIG="{prefix}conf/debug.conf"'
        )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "build -s samples/settings",
        expected_west_cmd=west_cmd_fmt(west_workplace_parametrized["prefix"]),
        should_succed=True,
    )


def test_sample_with_inherit_and_with_source_dir_and_board(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    Test if sample with inherit key and source_dir works, when there is board involved.
    """

    project_path = west_workplace_parametrized["project"]

    def west_cmd_fmt(prefix):
        return (
            "build -b nrf52840dk_nrf52840 samples/settings --"
            f" -DCONF_FILE={prefix}conf/common.conf"
            " -DOVERLAY_CONFIG="
            f'"{prefix}conf/nrf52840dk_nrf52840.conf;{prefix}conf/debug.conf"'
        )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "build -s samples/settings -b nrf52840dk_nrf52840",
        expected_west_cmd=west_cmd_fmt(west_workplace_parametrized["prefix"]),
        should_succed=True,
    )


east_yaml_no_apps_key = """
samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
"""


def test_no_apps_key_in_east_yml_build_type(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    Apps key is optional.
    Running east build with no apps key in east.yml should fail if --build-type is
    given.
    """

    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_apps_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build --build-type debug",
        should_succed=False,
    )


def test_no_apps_key_in_east_yml_app(west_workplace_parametrized, monkeypatch, mocker):
    """
    Apps key is optional.
    Running east build with no apps key in east.yml and building for app should not emit
    any extra cmake args.
    """

    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_apps_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build",
        "build",
    )


def test_no_apps_key_in_east_yml_sample(
    west_workplace_parametrized, monkeypatch, mocker
):
    """
    Apps key is optional.
    Running east build with no apps key in east.yml should not emit any extra cmake
    args.
    """

    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_apps_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        os.path.join(west_workplace_parametrized["project"], "samples", "settings"),
        "build",
        "build",
        should_succed=True,
    )


east_yaml_no_samples_key = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
"""


def test_no_samples_key_in_east_yml(west_workplace_parametrized, monkeypatch, mocker):
    """
    Samples key is optional.
    Running east build with no samples key in east.yml in samples should not emit any
    extra args.
    """

    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_samples_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        os.path.join(west_workplace_parametrized["project"], "samples", "settings"),
        "build",
        "build",
        should_succed=True,
    )


east_yaml_empty_apps_key = """
apps:
"""


def test_empty_apps_key(west_workplace_parametrized, monkeypatch, mocker):
    """
    Empty apps key is not allowed.
    """

    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_empty_apps_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build",
        should_succed=False,
    )


@pytest.mark.parametrize(
    "build_path",
    [
        "project",
        "test",
    ],
)
def test_building_outside_of_app_and_samples(
    west_workplace_parametrized, build_path, monkeypatch, mocker
):
    """
    Empty apps key is not allowed, and samples can not inherit from it.
    """

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized[build_path],
        "build -b native_posix",
        "build -b native_posix",
        should_succed=True,
    )



from click.testing import CliRunner
from setuptools_scm import get_version

from east.__main__ import cli


def test_version():
    runner = CliRunner()
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0

    assert (
        get_version(version_scheme="post-release", local_scheme="no-local-version")
        in result.output
    )



import shutil

import click

from ..east_context import east_command_settings, east_group_settings


@click.command(**east_command_settings)
@click.pass_obj
def env(east):
    """Update env"""
    pass


ncs_installed_msg = """Correct version of [bold cyan]NCS[/] toolchain is [bold]already[/] installed!

You can reinstall it with [bold cyan]--force[/] flag.
"""


@click.command(**east_command_settings)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Reinstall the toolchain, even if it is already installed.",
)
@click.pass_obj
def toolchain(east, force):
    """Update NCS toolchain for the current West workspace.

    Determines the version of the NCS that is being used in the current West
    workspace and downloads the correct version of the toolchain for it.

    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """

    east.pre_workspace_command_check(
        ignore_uninstalled_ncs=True,
        ignore_unsupported_ncs=False,
    )

    # We come to here if the detected ncs version is supported
    if east.ncs_version_installed:
        if force:
            shutil.rmtree(
                f"{east.consts['east_dir']}/toolchain/{east.detected_ncs_version}"
            )
        else:
            east.print(ncs_installed_msg)
            east.exit()

    east.print(
        "Starting install of [bold cyan]NCS[/] toolchain,"
        f" version [bold]{east.detected_ncs_version}[/], this will take some time...\n",
        highlight=False,
    )
    east.run_manager(f"install {east.detected_ncs_version}")

    east.print("\n[bold green]Done!")


@click.command(**east_command_settings)
@click.pass_obj
def west(east):
    """Update west"""
    pass


@click.group(**east_group_settings, subcommand_metavar="Subcommands")
@click.pass_obj
def update(east):
    """Command with several subcommands related to updating things."""
    pass


# update.add_command(env)
update.add_command(toolchain)
# update.add_command(west)
